/*
   Copyright 2008-2009 Murali Krishnan

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

package com.notuvy.net.webcache;

import java.io.File;
import java.util.*;
import java.text.SimpleDateFormat;

import com.notuvy.util.WallClock;
import org.jdom.Element;

import com.notuvy.file.Directory;
import com.notuvy.xml.FileStorage;
import com.notuvy.xml.SimpleJdom;

/**
 * A caching web proxy.
 * For simple get requests, the content is fetched and cached in a file.
 * The timeout of the cache can be customized.
 * Because the local file system is used for storage, there is no enforced size bound.
 * A background thread is needed to cull expired items so that disk usage does not grow continuously.
 *
 * User: murali
 * Date: Sep 9, 2008
 * Time: 5:19:30 PM
 */
public class CachedWebProxy extends SimpleJdom {

    //------------------------------------------------------------
    //- Class Variables

    protected static final String CACHE_SUBDIR = "download_cache";
    private static final String CACHE_FILENAME = "download_cache.xml";

    protected static final int NO_EXPIRATION = -1;

    protected static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("yyyy-MMM-dd HH:mm:ss");

    protected static final String XML_NODE = "Cache";

    // TODO: Create a background thread to cull expired items.

    //------------------------------------------------------------
    //- Class Functions

    //------------------------------------------------------------
    //- Instance Variables

    private final Map<String, CachedItem> fMap = new HashMap<String,CachedItem>();
    private final Directory fItemsDir;
    private long fExpiration = NO_EXPIRATION;

    //------------------------------------------------------------
    //- Constructors

    /**
     * Create the caching web proxy with storage in a subdirectory under the given directory.
     *
     * @param pParent  The directory under which the storage subdirectory is managed.
     */
    public CachedWebProxy(File pParent) {
        setStorage(new FileStorage(this, new File(pParent, CACHE_FILENAME)));
        fItemsDir = Directory.ensure(new File(pParent, CACHE_SUBDIR));

        getStorage().readOrCreate();

        removeOrphans();
    }

    //------------------------------------------------------------
    //- Getters

    public long getExpiration() {
        return fExpiration;
    }

    //------------------------------------------------------------
    //- Settors

    /**
     * Define the default expiration time for all queries.
     *
     * @param pExpiration  Query expiration time in milliseconds.
     */
    public void setExpiration(long pExpiration) {
        fExpiration = pExpiration;
    }

    //------------------------------------------------------------
    //- Private/Protected Utility Functions

    protected void removeOrphans() {
        Set<File> referenced = new HashSet<File>();
        for (CachedItem item : fMap.values()) {
            referenced.add(item.getFile());
        }

        for (File ondisk : fItemsDir.listContents()) {
            if (!referenced.contains(ondisk)) {
                LOG.warn("Removing orphan [" + ondisk.getName() + "]");
                if (!ondisk.delete()) {
                    LOG.error(String.format("Could not delete [%s].", ondisk.getAbsolutePath()));
                }
            }
        }
    }

    private CachedItem lookupItem(CacheQuery pQuery) {
        // todo: url normalization?
        Date now = WallClock.currentDate();

        CachedItem item = fMap.get(pQuery.getUrl());
        if (item == null) {
            item = new CachedItem(fItemsDir.getFile(), pQuery.getUrl(), now, pQuery.getExpiration());
            LOG.debug(String.format("Downloading first [%s].", pQuery.getUrl()));
            item.download();
            fMap.put(pQuery.getUrl(), item);
        } else {
            // Do not just use the background expiration time.  That was calculated from and older
            // query and may not be relevant.  Calculate the expiration using the value in this query.
            long expirationTime = pQuery.computeExpirationTime(item.getCreated().getTime());
            boolean redownloadNeeded = expirationTime <= now.getTime();
            // If an expiration was set and it has expired, then re-download.
            if (redownloadNeeded) {
                LOG.debug(String.format("Redownloading [%s].", pQuery.getUrl()));
                item.download();
                item.setCreated(now);
            } else {
                LOG.debug(String.format("Serve cached[%s]; created[%s].", pQuery.getUrl(), item.getCreated()));
            }
            item.setLastAccessed(now);
            item.calculateBackgroundExpiration(pQuery.getExpiration());
        }
        getStorage().save();  // At the very least, the timestamp changed.
        return item;
    }

    protected void printSummary() {
        List<CachedItem> items = listItems(CachedItem.BY_CREATION);

        StringBuilder builder = new StringBuilder("Cached Items:\n");
        for (CachedItem item : items) {
            builder.append("  [").append(DATE_FORMAT.format(item.getCreated())).append("] ").append(item.getUrlString()).append("\n");
        }

        LOG.info(builder);
    }

    protected List<CachedItem> listItems(Comparator<CachedItem> pSortOrder) {
        List<CachedItem> items = new ArrayList<CachedItem>(fMap.size());
        items.addAll(fMap.values());
        Collections.sort(items, pSortOrder);
        return items;
    }

    //------------------------------------------------------------
    //- Public Interface Functions

    public int size() {
        return fMap.size();
    }

    @Deprecated
    public File lookupFile(String pUrl) {
        return createQuery(pUrl).lookup().getFile();
    }

    @Deprecated
    public String lookupString(String pUrl) {
        return createQuery(pUrl).lookup().getStringContent();
    }

    public synchronized CacheResult lookup(CacheQuery pQuery) {
        CachedItem item = lookupItem(pQuery);
        return new CacheResult(item.getFile());
    }

    public CacheQuery createQuery(String pUrl) {
        return new CacheQuery(this, pUrl, getExpiration());
    }

    //------------------------------------------------------------
    //- Class Interface Functions

    public Element toXml() {
        Element result = new Element(XML_NODE);
        for (CachedItem item : fMap.values()) {
            result.addContent(item.toXml());
        }
        return result;
    }

    protected void parse(Element pRoot) {
        for (Object object : pRoot.getChildren(CachedItem.XML_NODE)) {
            CachedItem item = new CachedItem(fItemsDir.getFile());
            item.parse((Element) object);
            if (item.getFile().exists()) {
                fMap.put(item.getUrlString(), item);
            } else {
                LOG.warn("Missing content file [" + item.getFile().getName() + "]");
            }
        }
    }

    //------------------------------------------------------------
    //- Inner Classes

    //------------------------------------------------------------
    //- Main

    public static void main(String[] args) {
        for (String arg : args) {
            File parent = new File(arg);
            File actual = new File(parent, CACHE_SUBDIR);
            if (actual.exists()) {
                CachedWebProxy cd = new CachedWebProxy(parent);
                cd.printSummary();
            } else {
                LOG.error("No such directory [" + actual.getAbsolutePath() + "]");
            }
        }
    }
}
