/*
   Copyright 2008-2009 Murali Krishnan

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */

package com.notuvy.xml;


import org.apache.log4j.NDC;

import java.io.*;

/**
 * Save the XML to the file system.
 *
 * User: murali
 * Date: Dec 30, 2007
 * Time: 7:56:40 AM
 */
public class FileStorage extends BaseStorage {

    //------------------------------------------------------------
    //- Class Variables


    //------------------------------------------------------------
    //- Class Functions

    //------------------------------------------------------------
    //- Instance Variables

    private final File fFileImage;


    //------------------------------------------------------------
    //- Constructors

    public FileStorage(SimpleJdom pObject, File pFileImage) {
        super(pObject);
        fFileImage = pFileImage;
    }

    //------------------------------------------------------------
    //- Accessors

    public File getFileImage() {
        return fFileImage;
    }


    //------------------------------------------------------------
    //- Settors


    //------------------------------------------------------------
    //- Private/Protected Utility Functions

    private XmlIOStatus<Void> saveToStorage() {
        XmlIOStatus<Void> status = XmlIOStatus.failure();  // Pessimistic.
        try {
            FileWriter writer = new FileWriter(getFileImage());
            XmlIOStatus<Void> intermediate = getObject().saveTo(writer);
            if (!intermediate.isSuccess()) {
                LOG.error("Problem writing to [" + getFileImage().getAbsolutePath() + "]");
            }
            writer.close();

            takeContentSnapshot();

            status = intermediate;
        } catch (IOException ioe) {
            LOG.error("Cannot write to [" + getFileImage().getAbsolutePath() + "]", ioe);
        }
        return status;
    }


    //------------------------------------------------------------
    //- Public Interface Functions

    public XmlIOStatus<Void> readIfStoredImageExists(boolean pVerboseWarn) {
        XmlIOStatus<Void> status = XmlIOStatus.success();  // Vacuously true if no file exists.
        if (getFileImage().exists()) {
            status = read();
        } else if (pVerboseWarn) {
            LOG.warn(String.format("No input file [%s].", getFileImage().getAbsolutePath()));
        }
        return status;
    }


    //------------------------------------------------------------
    //- Class Interface Functions

    public XmlIOStatus<Void> save() {
        return skipSave() ? XmlIOStatus.success() : saveToStorage();
    }

    public XmlIOStatus<Void> read() {
        XmlIOStatus<Void> status = XmlIOStatus.failure();  // Pessimistic.
        NDC.push(String.format(" [%s]", getFileImage().getAbsolutePath()));
        try {
            FileReader reader = new FileReader(getFileImage());
            status = getObject().readFrom(reader);
            reader.close();

            if (status.isSuccess()) {
                takeContentSnapshot();
            }
        } catch (FileNotFoundException fnfe) {
            LOG.error("Cannot read from [" + getFileImage().getAbsolutePath() + "]", fnfe);
        } catch (IOException ioe) {
            LOG.error("Data read problem.", ioe);
        } finally {
            NDC.pop();
        }
        return status;
    }

    public XmlIOStatus<Boolean> exists() {
        return new XmlIOStatus<Boolean>(getFileImage().exists());
    }

    public XmlIOStatus<Void> readIfStoredImageExists() {
        return readIfStoredImageExists(true);
    }

    public XmlIOStatus<Void> readOrCreate() {
        XmlIOStatus<Void> status;
        if (getFileImage().exists()) {
            status = read();
        } else {
            status = saveToStorage();
            LOG.warn("Created: " + getFileImage().getAbsolutePath());
        }
        return status;
    }

    public XmlIOStatus<Void> delete() {
        XmlIOStatus<Void> status = XmlIOStatus.success();  // Optimistic.
        if (getFileImage().exists() && !getFileImage().delete()) {
            status = XmlIOStatus.failure(String.format("Could not delete [%s].", getFileImage().getAbsolutePath()));
            LOG.warn(status.getErrorMessage());
        }
        return status;
    }


    //------------------------------------------------------------
    //- Inner Classes

    //------------------------------------------------------------
    //- Main

}
