/*
   Copyright 2008-2009 Murali Krishnan

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */

package com.notuvy.xml;


import org.apache.log4j.NDC;

import java.io.*;

/**
 * A generic object that can be represented as XML and saved to a file.
 * This is useful when the full DOM tree is small enough to be comfortably
 * held all in memory.
 * 
 * User: murali
 * Date: Dec 30, 2007
 * Time: 7:56:40 AM
 */
public abstract class PersistentJdom extends SimpleJdom {

    //------------------------------------------------------------
    //- Class Variables

    //------------------------------------------------------------
    //- Class Functions

    //------------------------------------------------------------
    //- Instance Variables

    private final File fileImage;
    private boolean fOnlySaveIfChanged = false;
    private String fContentSnapshot = "";


    //------------------------------------------------------------
    //- Constructors

    protected PersistentJdom(File pFileImage) {
        fileImage = pFileImage;
    }

    //------------------------------------------------------------
    //- Accessors

    public File getFileImage() {
        return fileImage;
    }

    public boolean isOnlySaveIfChanged() {
        return fOnlySaveIfChanged;
    }


    //------------------------------------------------------------
    //- Settors

    public void setOnlySaveIfChanged(boolean pOnlySaveIfChanged) {
        fOnlySaveIfChanged = pOnlySaveIfChanged;
    }


    //------------------------------------------------------------
    //- Private/Protected Utility Functions

    /**
     * There is one situatin where we do a NOOP.  That is if we are checking for
     * changed content, and the contents have not changed.  In that case just return
     * 'success'.  Otherwise return the status of the save operation.
     *
     * @return boolean flag whether the actual save operation should be skipped.
     */
    private boolean skipSave() {
        return isOnlySaveIfChanged() && toXmlString().equals(fContentSnapshot);
    }

    private XmlIOStatus saveToFileImage() {
        XmlIOStatus success = XmlIOStatus.failure();  // Pessimistic.
        try {
            FileWriter writer = new FileWriter(getFileImage());
            XmlIOStatus intermediate = saveTo(writer);
            if (!intermediate.isSuccess()) {
                LOG.error("Problem writing to [" + getFileImage().getAbsolutePath() + "]");
            }
            writer.close();

            takeContentSnapshot();

            success = intermediate;
        } catch (IOException ioe) {
            LOG.error("Cannot write to [" + getFileImage().getAbsolutePath() + "]", ioe);
        }
        return success;
    }

    private void takeContentSnapshot() {
        if (isOnlySaveIfChanged()) {
            fContentSnapshot = toXmlString();
        }
    }


    //------------------------------------------------------------
    //- Public Interface Functions

    public XmlIOStatus save() {
        return skipSave() ? XmlIOStatus.success() : saveToFileImage();
    }

    public XmlIOStatus read() {
        XmlIOStatus status = XmlIOStatus.failure();  // Pessimistic.
        NDC.push(String.format(" [%s]", getFileImage().getAbsolutePath()));
        try {
            FileReader reader = new FileReader(getFileImage());
            status = readFrom(reader);
            reader.close();

            if (status.isSuccess()) {
                takeContentSnapshot();
            }
        } catch (FileNotFoundException fnfe) {
            LOG.error("Cannot read from [" + getFileImage().getAbsolutePath() + "]", fnfe);
        } catch (IOException ioe) {
            LOG.error("Data read problem.", ioe);
        } finally {
            NDC.pop();
        }
        return status;
    }

    public XmlIOStatus readIfFileExists() {
        XmlIOStatus status = XmlIOStatus.success();  // Vacuously true if no file exists.
        if (getFileImage().exists()) {
            status = read();
        }
        return status;
    }

    public XmlIOStatus readOrCreate() {
        XmlIOStatus status;
        if (getFileImage().exists()) {
            status = read();
        } else {
            status = saveToFileImage();
            LOG.warn("Created: " + getFileImage().getAbsolutePath());
        }
        return status;
    }


    //------------------------------------------------------------
    //- Class Interface Functions

    //------------------------------------------------------------
    //- Inner Classes

    //------------------------------------------------------------
    //- Main

}
