package com.notuvy.xml;


/**
 * The result of an IO operation.
 * Instead of returning a simple boolean success flag, IO operation can return this object
 * which encapsulates the boolean flag, plus also holds a descriptive error message for
 * the failure case.
 *
 * If the operation produces a return value, it can be returned in the payload.
 *
 * @author murali
 */
public class XmlIOStatus<T> {

    //------------------------------------------------------------
    //- Class Variables

    private static final XmlIOStatus<Void> SUCCESS = new XmlIOStatus<Void>(true, "");
    private static final XmlIOStatus<Void> UNSPECIFIED = new XmlIOStatus<Void>(false, "Unspecified error.");


    //------------------------------------------------------------
    //- Class Functions

    public static XmlIOStatus<Void> success() {
        return SUCCESS;
    }

    public static XmlIOStatus<Void> failure(String pErrorMessage) {
        return new XmlIOStatus<Void>(false, pErrorMessage);
    }

    public static XmlIOStatus<Void> failure() {
        return UNSPECIFIED;
    }


    //------------------------------------------------------------
    //- Instance Variables

    private final boolean fSuccess;
    private final String fErrorMessage;
    private final T fValue;


    //------------------------------------------------------------
    //- Constructors

    private XmlIOStatus(boolean pSuccess, String pErrorMessage) {
        this(pSuccess, pErrorMessage, null);
    }

    private XmlIOStatus(boolean pSuccess, String pErrorMessage, T pValue) {
        fSuccess = pSuccess;
        fErrorMessage = pErrorMessage;
        fValue = pValue;
    }

    /**
     * Create a successful response with the given payload.
     *
     * @param pValue The return value of the operation.
     */
    public XmlIOStatus(T pValue) {
        this (true, "", pValue);
    }

    /**
     * Create a failure response with the given payload.
     *
     * @param pErrorMessage  The description of the error.
     * @param pValue The return value of the operation.
     */
    public XmlIOStatus(String pErrorMessage, T pValue) {
        this (false, pErrorMessage, pValue);
    }


    //------------------------------------------------------------
    //- Gettors

    public boolean isSuccess() {
        return fSuccess;
    }

    public String getErrorMessage() {
        return fErrorMessage;
    }

    public T getValue() {
        return fValue;
    }


    //------------------------------------------------------------
    //- Settors


    //------------------------------------------------------------
    //- Private/Protected Utility Functions


    //------------------------------------------------------------
    //- Public Interface Functions


    //------------------------------------------------------------
    //- Class Interface Functions


    //------------------------------------------------------------
    //- Inner Classes


    //------------------------------------------------------------
    //- Main


}
